@@ -1,29 +1,25 @@
-#!/usr/bin/python
+#!/usr/bin/env python
 # -*- coding: utf8 -*-
 #
 # library for communicating with an isc dhcp server over the omapi protocol
 #
-# Copyright (C) 2010-2013 Cygnus Networks GmbH
+# Copyright 2010-2015 Cygnus Networks GmbH
 #
-# This program is free software: you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
-# the Free Software Foundation, either version 3 of the License, or
-# (at your option) any later version.
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
 #
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
+#     http://www.apache.org/licenses/LICENSE-2.0
 #
-# You should have received a copy of the GNU General Public License
-# along with this program.  If not, see <http://www.gnu.org/licenses/>.
-#
-"""
-For an example see http://code.google.com/p/pypureomapi/wiki/Example.
-"""
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
+
 
 # Message format:
-# 
+
 # authid (netint32)
 # authlen (netint32)
 # opcode (netint32)
@@ -33,16 +29,17 @@ For an example see http://code.google.com/p/pypureomapi/wiki/Example.
 # message (dictionary)
 # object (dictionary)
 # signature (length is authlen)
-# 
+
 # dictionary = entry* 0x00 0x00
 # entry = key (net16str) value (net32str)
 
-__author__      = "Helmut Grohne, Torge Szczepanek"
-__copyright__   = "Cygnus Networks GmbH"
-__licence__     = "GPL-3"
-__version__     = "0.3"
-__maintainer__  = "Torge Szczepanek"
-__email__       = "info@cygnusnetworks.de"
+__author__ = "Helmut Grohne, Dr. Torge Szczepanek"
+__copyright__ = "Cygnus Networks GmbH"
+__license__ = "Apache-2.0"
+__version__ = "0.4"
+__maintainer__ = "Dr. Torge Szczepanek"
+__email__ = "debian@cygnusnetworks.de"
+
 
 __all__ = []
 
@@ -57,48 +54,56 @@ import operator
 try:
 	basestring
 except NameError:
-	basestring = str
+	basestring = str  # pylint:disable=W0622
 
 logger = logging.getLogger("pypureomapi")
 sysrand = random.SystemRandom()
 
 __all__.extend("OMAPI_OP_OPEN OMAPI_OP_REFRESH OMAPI_OP_UPDATE".split())
 __all__.extend("OMAPI_OP_NOTIFY OMAPI_OP_STATUS OMAPI_OP_DELETE".split())
-OMAPI_OP_OPEN    = 1
+OMAPI_OP_OPEN = 1
 OMAPI_OP_REFRESH = 2
-OMAPI_OP_UPDATE  = 3
-OMAPI_OP_NOTIFY  = 4
-OMAPI_OP_STATUS  = 5
-OMAPI_OP_DELETE  = 6
+OMAPI_OP_UPDATE = 3
+OMAPI_OP_NOTIFY = 4
+OMAPI_OP_STATUS = 5
+OMAPI_OP_DELETE = 6
+
 
 def repr_opcode(opcode):
 	"""Returns a textual representation for the given opcode.
 	@type opcode: int
 	@rtype: str
 	"""
-	opmap = {1: "open", 2: "refresh", 3: "update", 4: "notify", 5: "status",
-			6: "delete"}
+	opmap = {1: "open", 2: "refresh", 3: "update", 4: "notify", 5: "status", 6: "delete"}
 	return opmap.get(opcode, "unknown (%d)" % opcode)
 
 __all__.append("OmapiError")
+
+
 class OmapiError(Exception):
 	"""OMAPI exception base class."""
 
 __all__.append("OmapiSizeLimitError")
+
+
 class OmapiSizeLimitError(OmapiError):
 	"""Packet size limit reached."""
 	def __init__(self):
 		OmapiError.__init__(self, "Packet size limit reached.")
 
 __all__.append("OmapiErrorNotFound")
+
+
 class OmapiErrorNotFound(OmapiError):
 	"""Not found."""
 	def __init__(self):
 		OmapiError.__init__(self, "not found")
 
+
 class OutBuffer(object):
 	"""Helper class for constructing network packets."""
 	sizelimit = 65536
+
 	def __init__(self):
 		self.buff = io.BytesIO()
 
@@ -191,7 +196,7 @@ class OutBuffer(object):
 			items = items.items()
 		for key, value in items:
 			self.add_net16string(key).add_net32string(value)
-		return self.add(b"\x00\x00") # end marker
+		return self.add(b"\x00\x00")  # end marker
 
 	def getvalue(self):
 		"""
@@ -213,6 +218,7 @@ class OutBuffer(object):
 		self.buff = io.BytesIO(self.getvalue()[length:])
 		return self
 
+
 class OmapiStartupMessage(object):
 	"""Class describing the protocol negotiation messages.
 
@@ -269,8 +275,8 @@ class OmapiStartupMessage(object):
 		@rtype: str
 		@returns: a human readable representation in one line
 		"""
-		return "protocol_version=%d header_size=%d" % (self.protocol_version,
-				self.header_size)
+		return "protocol_version=%d header_size=%d" % (self.protocol_version, self.header_size)
+
 
 class OmapiAuthenticatorBase(object):
 	"""Base class for OMAPI authenticators.
@@ -281,17 +287,20 @@ class OmapiAuthenticatorBase(object):
 	@ivar authid: is the authenticator id as assigned during the handshake
 	@type authid: int
 	"""
-	authlen = -1 # must be overwritten
+	authlen = -1  # must be overwritten
 	algorithm = None
-	authid = -1 # will be an instance attribute
+	authid = -1  # will be an instance attribute
+
 	def __init__(self):
 		pass
+
 	def auth_object(self):
 		"""
 		@rtype: {bytes: bytes}
 		@returns: object part of an omapi authentication message
 		"""
 		raise NotImplementedError
+
 	def sign(self, message):
 		"""
 		@type message: bytes
@@ -300,19 +309,25 @@ class OmapiAuthenticatorBase(object):
 		"""
 		raise NotImplementedError()
 
+
 class OmapiNullAuthenticator(OmapiAuthenticatorBase):
 	authlen = 0
-	authid = 0 # always 0
+	authid = 0  # always 0
+
 	def __init__(self):
 		OmapiAuthenticatorBase.__init__(self)
+
 	def auth_object(self):
 		return {}
+
 	def sign(self, _):
 		return b""
 
+
 class OmapiHMACMD5Authenticator(OmapiAuthenticatorBase):
 	authlen = 16
 	algorithm = b"hmac-md5.SIG-ALG.REG.INT."
+
 	def __init__(self, user, key):
 		"""
 		@type user: bytes
@@ -340,6 +355,8 @@ class OmapiHMACMD5Authenticator(OmapiAuthenticatorBase):
 		return hmac.HMAC(self.key, message).digest()
 
 __all__.append("OmapiMessage")
+
+
 class OmapiMessage(object):
 	"""
 	@type authid: int
@@ -361,8 +378,7 @@ class OmapiMessage(object):
 	@ivar signature: A signature on this message as generated by an
 			authenticator.
 	"""
-	def __init__(self, authid=0, opcode=0, handle=0, tid=0, rid=0,
-			message=None, obj=None, signature=b""):
+	def __init__(self, authid=0, opcode=0, handle=0, tid=0, rid=0, message=None, obj=None, signature=b""):
 		"""
 		Construct an OmapiMessage from the given fields. No error
 		checking is performed.
@@ -393,7 +409,7 @@ class OmapiMessage(object):
 		>>> OmapiMessage(tid=-1).tid != OmapiMessage(tid=-1).tid
 		True
 		"""
-		self.tid = sysrand.randrange(0, 1<<32)
+		self.tid = sysrand.randrange(0, 1 << 32)
 
 	def serialize(self, outbuffer, forsigning=False):
 		"""
@@ -431,7 +447,7 @@ class OmapiMessage(object):
 		@type authenticator: OmapiAuthenticatorBase
 		"""
 		self.authid = authenticator.authid
-		self.signature = b"\0" * authenticator.authlen # provide authlen
+		self.signature = b"\0" * authenticator.authlen  # provide authlen
 		self.signature = authenticator.sign(self.as_string(forsigning=True))
 		assert len(self.signature) == authenticator.authlen
 
@@ -455,9 +471,7 @@ class OmapiMessage(object):
 		@rtype: bool
 		"""
 		try:
-			return authenticators[self.authid]. \
-					sign(self.as_string(forsigning=True)) == \
-					self.signature
+			return authenticators[self.authid]. sign(self.as_string(forsigning=True)) == self.signature
 		except KeyError:
 			return False
 
@@ -496,8 +510,7 @@ class OmapiMessage(object):
 		"""
 		@type update: {bytes: bytes}
 		"""
-		self.obj = [(key, value) for key, value in self.obj
-					if key not in update]
+		self.obj = [(key, value) for key, value in self.obj if key not in update]
 		self.obj.extend(update.items())
 
 	def dump(self):
@@ -505,27 +518,15 @@ class OmapiMessage(object):
 		@rtype: str
 		@returns: a human readable representation of the message
 		"""
-		return "".join(("Omapi message attributes:\n",
-				"authid:\t\t%d\n" % self.authid,
-				"authlen:\t%d\n" % len(self.signature),
-				"opcode:\t\t%s\n" % repr_opcode(self.opcode),
-				"handle:\t\t%d\n" % self.handle,
-				"tid:\t\t%d\n" % self.tid,
-				"rid:\t\t%d\n" % self.rid,
-				"message:\t%r\n" % self.message,
-				"obj:\t\t%r\n" % self.obj,
-				"signature:\t%r\n" % self.signature))
+		return "".join(("Omapi message attributes:\n", "authid:\t\t%d\n" % self.authid, "authlen:\t%d\n" % len(self.signature), "opcode:\t\t%s\n" % repr_opcode(self.opcode), "handle:\t\t%d\n" % self.handle, "tid:\t\t%d\n" % self.tid, "rid:\t\t%d\n" % self.rid, "message:\t%r\n" % self.message, "obj:\t\t%r\n" % self.obj, "signature:\t%r\n" % self.signature))
 
 	def dump_oneline(self):
 		"""
 		@rtype: str
 		@returns: a barely human readable representation in one line
 		"""
-		return ("authid=%d authlen=%d opcode=%s handle=%d tid=%d rid=%d " +
-				"message=%r obj=%r signature=%r") % (self.authid,
-						len(self.signature), repr_opcode(self.opcode),
-						self.handle, self.tid, self.rid, self.message,
-						self.obj, self.signature)
+		return ("authid=%d authlen=%d opcode=%s handle=%d tid=%d rid=%d message=%r obj=%r signature=%r") % (self.authid, len(self.signature), repr_opcode(self.opcode), self.handle, self.tid, self.rid, self.message, self.obj, self.signature)
+
 
 def parse_map(filterfun, parser):
 	"""Creates a new parser that passes the result of the given parser through
@@ -545,6 +546,7 @@ def parse_map(filterfun, parser):
 			yield filterfun(element)
 			break
 
+
 def parse_chain(*args):
 	"""Creates a new parser that executes the passed parsers (args) with the
 	previous results and yields a tuple of the results.
@@ -565,8 +567,10 @@ def parse_chain(*args):
 				break
 	yield tuple(items)
 
+
 class InBuffer(object):
 	sizelimit = 65536
+
 	def __init__(self, initial=b""):
 		"""
 		@type initial: bytes
@@ -612,32 +616,28 @@ class InBuffer(object):
 		>>> hex(next(InBuffer(b"\\x01\\x02").parse_net16int()))
 		'0x102'
 		"""
-		return parse_map(lambda data: struct.unpack("!H", data)[0],
-				self.parse_fixedbuffer(2))
+		return parse_map(lambda data: struct.unpack("!H", data)[0], self.parse_fixedbuffer(2))
 
 	def parse_net32int(self):
 		"""
 		>>> hex(int(next(InBuffer(b"\\x01\\0\\0\\x02").parse_net32int())))
 		'0x1000002'
 		"""
-		return parse_map(lambda data: struct.unpack("!L", data)[0],
-				self.parse_fixedbuffer(4))
+		return parse_map(lambda data: struct.unpack("!L", data)[0], self.parse_fixedbuffer(4))
 
 	def parse_net16string(self):
 		"""
 		>>> next(InBuffer(b"\\0\\x03eggs").parse_net16string()) == b'egg'
 		True
 		"""
-		return parse_map(operator.itemgetter(1),
-				parse_chain(self.parse_net16int, self.parse_fixedbuffer))
+		return parse_map(operator.itemgetter(1), parse_chain(self.parse_net16int, self.parse_fixedbuffer))
 
 	def parse_net32string(self):
 		"""
 		>>> next(InBuffer(b"\\0\\0\\0\\x03eggs").parse_net32string()) == b'egg'
 		True
 		"""
-		return parse_map(operator.itemgetter(1),
-				parse_chain(self.parse_net32int, self.parse_fixedbuffer))
+		return parse_map(operator.itemgetter(1), parse_chain(self.parse_net32int, self.parse_fixedbuffer))
 
 	def parse_bindict(self):
 		"""
@@ -672,35 +672,36 @@ class InBuffer(object):
 		>>> d = b"\\0\\0\\0\\x64\\0\\0\\0\\x18"
 		>>> next(InBuffer(d).parse_startup_message()).validate()
 		"""
-		return parse_map(lambda args: OmapiStartupMessage(*args),
-				parse_chain(self.parse_net32int,
-					lambda _: self.parse_net32int()))
+		return parse_map(lambda args: OmapiStartupMessage(*args), parse_chain(self.parse_net32int, lambda _: self.parse_net32int()))
 
 	def parse_message(self):
 		"""results in an OmapiMessage"""
-		parser = parse_chain(self.parse_net32int, # authid
-				lambda *_: self.parse_net32int(), # authlen
-				lambda *_: self.parse_net32int(), # opcode
-				lambda *_: self.parse_net32int(), # handle
-				lambda *_: self.parse_net32int(), # tid
-				lambda *_: self.parse_net32int(), # rid
-				lambda *_: self.parse_bindict(), # message
-				lambda *_: self.parse_bindict(), # object
-				lambda *args: self.parse_fixedbuffer(args[1])) # signature
-		return parse_map(lambda args: # skip authlen in args:
+		parser = parse_chain(self.parse_net32int,  # authid
+							lambda *_: self.parse_net32int(),  # authlen
+							lambda *_: self.parse_net32int(),  # opcode
+							lambda *_: self.parse_net32int(),  # handle
+							lambda *_: self.parse_net32int(),  # tid
+							lambda *_: self.parse_net32int(),  # rid
+							lambda *_: self.parse_bindict(),  # message
+							lambda *_: self.parse_bindict(),  # object
+							lambda *args: self.parse_fixedbuffer(args[1]))  # signature
+		return parse_map(lambda args:  # skip authlen in args:
 				OmapiMessage(*(args[0:1] + args[2:])), parser)
 
 if isinstance(bytes(b"x")[0], int):
 	def bytes_to_int_seq(b):
 		return b
-	int_seq_to_bytes = bytes # raises ValueError
+	int_seq_to_bytes = bytes  # raises ValueError
 else:
 	def bytes_to_int_seq(b):
-		return map(ord, b)
+		return [ord(x) for x in b]
+
 	def int_seq_to_bytes(s):
-		return "".join(map(chr, s)) # raises ValueError
+		return "".join([chr(x) for x in s])  # raises ValueError
 
 __all__.append("pack_ip")
+
+
 def pack_ip(ipstr):
 	"""Converts an ip address given in dotted notation to a four byte
 	string in network byte order.
@@ -721,10 +722,12 @@ def pack_ip(ipstr):
 	parts = ipstr.split('.')
 	if len(parts) != 4:
 		raise ValueError("given ip address has an invalid number of dots")
-	parts = map(int, parts) # raises ValueError
-	return int_seq_to_bytes(parts) # raises ValueError
+	parts = [int(x) for x in parts]  # raises ValueError
+	return int_seq_to_bytes(parts)  # raises ValueError
 
 __all__.append("unpack_ip")
+
+
 def unpack_ip(fourbytes):
 	"""Converts an ip address given in a four byte string in network
 	byte order to a string in dotted notation.
@@ -744,9 +747,11 @@ def unpack_ip(fourbytes):
 		raise ValueError("given buffer is not a string")
 	if len(fourbytes) != 4:
 		raise ValueError("given buffer is not exactly four bytes long")
-	return ".".join(map(str, bytes_to_int_seq(fourbytes)))
+	return ".".join([str(x) for x in bytes_to_int_seq(fourbytes)])
 
 __all__.append("pack_mac")
+
+
 def pack_mac(macstr):
 	"""Converts a mac address given in colon delimited notation to a
 	six byte string in network byte order.
@@ -768,10 +773,12 @@ def pack_mac(macstr):
 	parts = macstr.split(":")
 	if len(parts) != 6:
 		raise ValueError("given mac addresses has an invalid number of colons")
-	parts = [int(part, 16) for part in parts] # raises ValueError
-	return int_seq_to_bytes(parts) # raises ValueError
+	parts = [int(part, 16) for part in parts]  # raises ValueError
+	return int_seq_to_bytes(parts)  # raises ValueError
 
 __all__.append("unpack_mac")
+
+
 def unpack_mac(sixbytes):
 	"""Converts a mac address given in a six byte string in network
 	byte order to a string in colon delimited notation.
@@ -791,15 +798,136 @@ def unpack_mac(sixbytes):
 		raise ValueError("given buffer is not a string")
 	if len(sixbytes) != 6:
 		raise ValueError("given buffer is not exactly six bytes long")
-	return ":".join(map("%2.2x".__mod__, bytes_to_int_seq(sixbytes)))
+	return ":".join(["%2.2x".__mod__(x) for x in bytes_to_int_seq(sixbytes)])
+
+
+class LazyStr(object):
 
-class lazy_str(object):
 	def __init__(self, function):
 		self.function = function
+
 	def __str__(self):
 		return self.function()
 
+
+class TCPClientTransport(object):
+	"""PEP 3156 dummy transport class to support OmapiProtocol class."""
+	def __init__(self, protocol, host, port):
+		self.protocol = protocol
+		self.connection = socket.socket()
+		self.connection.connect((host, port))
+		self.protocol.connection_made(self)
+
+	def close(self):
+		"""Close the omapi connection if it is open."""
+		if self.connection:
+			self.connection.close()
+			self.connection = None
+
+	def fill_inbuffer(self):
+		"""Read bytes from the connection and hand them to the protocol.
+		@raises OmapiError:
+		@raises socket.error:
+		"""
+		if not self.connection:
+			raise OmapiError("not connected")
+		try:
+			data = self.connection.recv(2048)
+		except socket.error:
+			self.close()
+			raise
+		if not data:
+			self.close()
+			raise OmapiError("connection closed")
+		try:
+			self.protocol.data_received(data)
+		except OmapiSizeLimitError:
+			self.close()
+			raise
+
+	def write(self, data):
+		"""Send all of data to the connection.
+
+		@type data: bytes
+		@raises socket.error:
+		"""
+		try:
+			self.connection.sendall(data)
+		except socket.error:
+			self.close()
+			raise
+
+
+class OmapiProtocol(object):
+	"""PEP 3156 like protocol class for Omapi.
+
+	This interface is not yet to be relied upon.
+	"""
+	def __init__(self):
+		self.transport = None
+		self.authenticators = {0: OmapiNullAuthenticator()}
+		self.defauth = 0
+		self.inbuffer = InBuffer()
+		self.current_parser = self.inbuffer.parse_startup_message()
+
+	def connection_made(self, transport):
+		self.transport = transport
+		message = OmapiStartupMessage()
+		logger.debug("sending omapi startup message %s", LazyStr(message.dump_oneline))
+		self.transport.write(message.as_string())
+
+	def data_received(self, data):
+		"""
+		@type data: bytes
+		"""
+		self.inbuffer.feed(data)
+		while True:
+			if self.current_parser is None:
+				self.current_parser = self.inbuffer.parse_message()
+			result = next(self.current_parser)
+			if result is None:
+				return
+			self.current_parser = None
+			self.inbuffer.resetsize()
+			if isinstance(result, OmapiStartupMessage):
+				logger.debug("received omapi startup message %s", LazyStr(result.dump_oneline))
+				self.startup_received(result)
+			else:
+				assert isinstance(result, OmapiMessage)
+				logger.debug("received %s", LazyStr(result.dump_oneline))
+				self.message_received(result)
+
+	def startup_received(self, startup_message):
+		try:
+			startup_message.validate()
+		except OmapiError:
+			self.transport.close()
+			raise
+		self.startup_completed()
+
+	@staticmethod
+	def startup_completed():
+		logger.debug("omapi connection initialized")
+
+	def message_received(self, message):
+		pass
+
+	def send_message(self, message, sign=True):
+		"""Send the given message to the connection.
+
+		@type message: OmapiMessage
+		@param sign: whether the message needs to be signed
+		@raises OmapiError:
+		@raises socket.error:
+		"""
+		if sign:
+			message.sign(self.authenticators[self.defauth])
+		logger.debug("sending %s", LazyStr(message.dump_oneline))
+		self.transport.write(message.as_string())
+
 __all__.append("Omapi")
+
+
 class Omapi(object):
 	def __init__(self, hostname, port, username=None, key=None):
 		"""
@@ -814,18 +942,17 @@ class Omapi(object):
 		"""
 		self.hostname = hostname
 		self.port = port
-		self.authenticators = {0: OmapiNullAuthenticator()}
-		self.defauth = 0
+		self.protocol = OmapiProtocol()
+		self.recv_message_queue = []
+		self.protocol.startup_completed = lambda: self.recv_message_queue.append(None)
+		self.protocol.message_received = self.recv_message_queue.append
 
 		newauth = None
 		if username is not None and key is not None:
 			newauth = OmapiHMACMD5Authenticator(username, key)
 
-		self.connection = socket.socket()
-		self.inbuffer = InBuffer()
-		self.connection.connect((hostname, port))
+		self.transport = TCPClientTransport(self.protocol, hostname, port)
 
-		self.send_protocol_initialization()
 		self.recv_protocol_initialization()
 
 		if newauth:
@@ -833,89 +960,24 @@ class Omapi(object):
 
 	def close(self):
 		"""Close the omapi connection if it is open."""
-		if self.connection:
-			self.connection.close()
-			self.connection = None
+		self.transport.close()
 
 	def check_connected(self):
 		"""Raise an OmapiError unless connected.
 		@raises OmapiError:
 		"""
-		if not self.connection:
+		if not self.transport.connection:
 			raise OmapiError("not connected")
 
-	def recv_conn(self, length):
-		"""Receive up to length bytes of data from the connection.
-
-		@type length: int
-		@rtype: bytes
-		@raises OmapiError: if not connected
-		@raises socket.error:
-		"""
-		self.check_connected()
-		try:
-			return self.connection.recv(length)
-		except socket.error:
-			self.close()
-			raise
-
-	def send_conn(self, data):
-		"""Send all of data to the connection.
-
-		@type data: bytes
-		@raises OmapiError: if not connected
-		@raises socket.error:
-		"""
-		self.check_connected()
-		try:
-			self.connection.sendall(data)
-		except socket.error:
-			self.close()
-			raise
-
-	def fill_inbuffer(self):
-		"""
-		@raises OmapiError:
-		@raises socket.error:
-		"""
-		data = self.recv_conn(2048)
-		if not data:
-			self.close()
-			raise OmapiError("connection closed")
-		try:
-			self.inbuffer.feed(data)
-		except OmapiSizeLimitError:
-			self.close()
-			raise
-
-	def send_protocol_initialization(self):
-		"""
-		@raises OmapiError:
-		@raises socket.error:
-		"""
-		self.check_connected()
-		message = OmapiStartupMessage()
-		logger.debug("sending omapi startup message %s",
-				lazy_str(message.dump_oneline))
-		self.send_conn(message.as_string())
-
 	def recv_protocol_initialization(self):
 		"""
 		@raises OmapiError:
 		@raises socket.error:
 		"""
-		for result in self.inbuffer.parse_startup_message():
-			if result is None:
-				self.fill_inbuffer()
-			else:
-				self.inbuffer.resetsize()
-				logger.debug("received omapi startup message %s",
-						lazy_str(result.dump_oneline))
-				try:
-					result.validate()
-				except OmapiError:
-					self.close()
-					raise
+		while not self.recv_message_queue:
+			self.transport.fill_inbuffer()
+		message = self.recv_message_queue.pop(0)
+		assert message is None
 
 	def receive_message(self):
 		"""Read the next message from the connection.
@@ -923,16 +985,14 @@ class Omapi(object):
 		@raises OmapiError:
 		@raises socket.error:
 		"""
-		for message in self.inbuffer.parse_message():
-			if message is None:
-				self.fill_inbuffer()
-			else:
-				self.inbuffer.resetsize()
-				logger.debug("received %s", lazy_str(message.dump_oneline))
-				if not message.verify(self.authenticators):
-					self.close()
-					raise OmapiError("bad omapi message signature")
-				return message
+		while not self.recv_message_queue:
+			self.transport.fill_inbuffer()
+		message = self.recv_message_queue.pop(0)
+		assert message is not None
+		if not message.verify(self.protocol.authenticators):
+			self.close()
+			raise OmapiError("bad omapi message signature")
+		return message
 
 	def receive_response(self, message, insecure=False):
 		"""Read the response for the given message.
@@ -947,9 +1007,8 @@ class Omapi(object):
 		if not response.is_response(message):
 			raise OmapiError("received message is not the desired response")
 		# signature already verified
-		if response.authid != self.defauth and not insecure:
-			raise OmapiError("received message is signed with wrong " +
-						"authenticator")
+		if response.authid != self.protocol.defauth and not insecure:
+			raise OmapiError("received message is signed with wrong authenticator")
 		return response
 
 	def send_message(self, message, sign=True):
@@ -961,10 +1020,7 @@ class Omapi(object):
 		@raises socket.error:
 		"""
 		self.check_connected()
-		if sign:
-			message.sign(self.authenticators[self.defauth])
-		logger.debug("sending %s", lazy_str(message.dump_oneline))
-		self.send_conn(message.as_string())
+		self.protocol.send_message(message, sign)
 
 	def query_server(self, message):
 		"""Send the message and receive a response for it.
@@ -975,7 +1031,6 @@ class Omapi(object):
 		"""
 		self.send_message(message)
 		return self.receive_response(message)
-		
 
 	def initialize_authenticator(self, authenticator):
 		"""
@@ -991,9 +1046,9 @@ class Omapi(object):
 		authid = response.handle
 		if authid == 0:
 			raise OmapiError("received invalid authid from server")
-		self.authenticators[authid] = authenticator
+		self.protocol.authenticators[authid] = authenticator
 		authenticator.authid = authid
-		self.defauth = authid
+		self.protocol.defauth = authid
 		logger.debug("successfully initialized default authid %d", authid)
 
 	def add_host(self, ip, mac):
@@ -1054,7 +1109,7 @@ class Omapi(object):
 			raise OmapiErrorNotFound()
 		try:
 			return unpack_ip(dict(response.obj)[b"ip-address"])
-		except KeyError: # ip-address
+		except KeyError:  # ip-address
 			raise OmapiErrorNotFound()
 
 	def lookup_mac(self, ip):
@@ -1076,9 +1131,31 @@ class Omapi(object):
 			raise OmapiErrorNotFound()
 		try:
 			return unpack_mac(dict(response.obj)[b"hardware-address"])
-		except KeyError: # hardware-address
+		except KeyError:  # hardware-address
 			raise OmapiErrorNotFound()
 
+	def add_host_supersede_name(omapi, ip, mac, name):  # pylint:disable=E0213
+		"""Add a host with a fixed-address and override its hostname with the given name.
+		@type omapi: Omapi
+		@type ip: str
+		@type mac: str
+		@type name: str
+		@raises ValueError:
+		@raises OmapiError:
+		@raises socket.error:
+		"""
+		msg = OmapiMessage.open(b"host")
+		msg.message.append((b"create", struct.pack("!I", 1)))
+		msg.message.append((b"exclusive", struct.pack("!I", 1)))
+		msg.obj.append((b"hardware-address", pack_mac(mac)))
+		msg.obj.append((b"hardware-type", struct.pack("!I", 1)))
+		msg.obj.append((b"ip-address", pack_ip(ip)))
+		msg.obj.append((b"name", name))
+		msg.obj.append((b"statements", str.encode('supersede host-name "%s";' % name)))
+		response = omapi.query_server(msg)
+		if response.opcode != OMAPI_OP_UPDATE:
+			raise OmapiError("add failed")
+
 if __name__ == '__main__':
 	import doctest
 	doctest.testmod()
